/******************************************************************************************************/
/* The 1st IPSJ SIG-ARC High-Performance Processor Design Contest                     Version 2013-10 */
/* from MipsCore pipeline95 2013-10-13                                            ArchLab. TOKYO TECH */
/******************************************************************************************************/
`include "../rtl/define.v"
`default_nettype none

/******************************************************************************************************/
/* 32bit-32cycle divider (signed or unsigned), Just for Pipelining Version                            */
/******************************************************************************************************/
module DIVUNIT(CLK, RST_X, INIT, SIGNED, A, B, RSLT, BUSY);
    input         CLK, RST_X, INIT, SIGNED;
    input  [31:0] A, B;
    output [63:0] RSLT;
    output        BUSY;

    wire [31:0]   uint_a = (SIGNED & A[31]) ? ~A + 1 : A;
    wire [31:0]   uint_b = (SIGNED & B[31]) ? ~B + 1 : B;
    wire [63:0]   uint_rslt;
    reg           sign, sign_a, sign_b;

    reg           init_t;
    always @(posedge CLK) init_t <= INIT;
    
    DIVUNITCORE divcore(.CLK(CLK), .RST_X(RST_X), .INIT(INIT), .A(uint_a), .B(uint_b), 
                        .RSLT(uint_rslt), .BUSY(BUSY));
   
    assign RSLT[63:32] = (~sign) ?                      uint_rslt[63:32]     :
                         ({sign_a, sign_b} == 2'b00) ?  uint_rslt[63:32]     :
                         ({sign_a, sign_b} == 2'b01) ?  uint_rslt[63:32]     :
                         ({sign_a, sign_b} == 2'b10) ? ~uint_rslt[63:32] + 1 : ~uint_rslt[63:32] + 1;

    assign RSLT[31: 0] = (~sign) ?                      uint_rslt[31: 0]     :
                         ({sign_a, sign_b} == 2'b00) ?  uint_rslt[31: 0]     :
                         ({sign_a, sign_b} == 2'b01) ? ~uint_rslt[31: 0] + 1 :
                         ({sign_a, sign_b} == 2'b10) ? ~uint_rslt[31: 0] + 1 : uint_rslt[31: 0];
        
    always @( posedge CLK or negedge RST_X )
      if(!RST_X) {sign, sign_a, sign_b} <= 0;
      else if (init_t==1 && INIT==0) begin
          sign    <= SIGNED;
          sign_a  <= A[31];
          sign_b  <= B[31];
      end
endmodule

/******************************************************************************************************/
module DIVUNITCORE(CLK, RST_X, INIT, A, B, RSLT, BUSY);
    input             CLK, RST_X, INIT;
    input      [31:0] A, B;
    output reg [63:0] RSLT;
    output            BUSY;

    reg [31:0]    divisor;
    reg [5:0]     count;
    wire [32:0]   differ = RSLT[63:31] - {1'b0, divisor};

    reg           init_t;
    always @(posedge CLK) init_t <= INIT;

    assign BUSY = count!=0 || (init_t==1 && INIT==0);
    
    always @(posedge CLK or negedge RST_X) begin
        if(!RST_X) {divisor, count, RSLT} <= 0;
        else begin
            if(init_t==1 && INIT==0) begin
                divisor <= B;
                RSLT    <= {32'h0, A};
                count   <= 1;
            end else if(count!=0 && count<33) begin
                divisor <= divisor;
                RSLT    <= (differ[32]) ? {RSLT[62:0], 1'h0} : {differ[31:0], RSLT[30:0], 1'h1};
                count   <= count + 1;
            end else begin
                count   <= 0;
            end
        end
    end
endmodule

/******************************************************************************************************/
/* MipsCore: 32bit-MIPS 5-stage pipelining processor                                                  */
/******************************************************************************************************/
module MIPSCORE(CLK, RST_X, STALL, STAT, I_ADDR, I_IN, D_ADDR, D_IN, D_OUT, D_OE, D_WE, IO_IN);
    input               CLK, RST_X, STALL;
    output reg [2:0]    STAT;
    output [`ADDR]      I_ADDR, D_ADDR;
    input  [31:0]       I_IN, D_IN, IO_IN;
    output [31:0]       D_OUT;
    output [3:0]        D_WE;
    output              D_OE;

    reg [`ADDR]         pc;        // program counter
    reg [31:0]          hi, lo;    // high and low register for multiply and divide inst.

    assign I_ADDR = pc;

    /**************************************************************************************************/
    reg [`ADDR] IfId_npc;     // IF-ID pipeline reg: next pc, pc + 4
    reg [31:0]  IfId_ir;      // IF-ID pipeline reg: instruction

    reg [`ADDR] IdEx_npc;     // ID-EX pipeline reg: next pc, pc + 4
    reg [31:0]  IdEx_rrs;     // ID-EX pipeline reg: fetched operand of R[rs]  
    reg [31:0]  IdEx_rrt;     // ID-EX pipeline reg: fetched operand of R[rt]
    reg [4:0]   IdEx_dst;     // ID-EX pipeline reg: decoded destination reg number
    reg [31:0]  IdEx_ir;      // ID-EX pipeline reg: instruction
    reg [`OPNT] IdEx_opn;     // ID-EX pipeline reg: decoded operation ID
    reg [`ATTR] IdEx_attr;    // ID-EX pipeline reg: decoded instruction attribute 

    reg [4:0]   ExMa_dst;     // EX-MA pipeline reg: decoded destination reg number
    reg [31:0]  ExMa_rslt;    // EX-MA pipeline reg: execution result
    reg [3:0]   ExMa_mwe;     // EX-MA pipeline reg: mem write enable
    reg         ExMa_oe;      // EX-MA pipeline reg: data memory output enable
    reg [4:0]   ExMa_lds;     // EX-MA pipeline reg: load selector
    reg [31:0]  ExMa_std;     // EX-MA pipeline reg: store data
    
    reg [4:0]   MaWb_dst;     // MA-WB pipeline reg: decoded destination reg number 
    reg [31:0]  MaWb_rslt;    // MA-WB pipeline reg: execution result

    /**************************************************************************************************/
    reg [`ADDR] IdTPC;        // wire, calculated branch Taken address (Taken Program Counter)
    reg         IdC;          // wire, Calculated branch result, branch Taken or Untaken
    reg         IdB;          // wire, branch/jump instruction ?
    wire [31:0] MaRSLT;       // wire, execution result on MA stage
    wire        bstall;       // stall signal by branch instruction
    wire        DIVBUSY;
    wire        PSTALL = (STALL || DIVBUSY);  // pipeline stall
    
    /**************************************************************************************************/
    /* Stage 1 : IF, instruction fetch                                                                */
    /**************************************************************************************************/
    wire [`ADDR] IfNPC = pc + 4;
    
    always @( posedge CLK or negedge RST_X ) begin ///// update program counter
        if(!RST_X)                  pc <= 0;
        else if(!PSTALL && !bstall) pc <= (IdC) ? IdTPC : IfNPC; // If branch taken, uses taken PC
    end

    always @( posedge CLK or negedge RST_X ) begin  ///// update pipeline registers
        if(!RST_X) {IfId_npc, IfId_ir} <= 0;
        else if(!PSTALL && !bstall) begin
            IfId_npc <= IfNPC;
            IfId_ir  <= I_IN;  // if branch taken then NOP else instruction from imem.
        end
    end
        
    /**************************************************************************************************/
    /* Stage 2 : ID, instruction decode & operand fetch                                               */
    /**************************************************************************************************/
    wire [5:0] IdOP   = IfId_ir[31:26]; // opcode field of instruction  
    wire [4:0] IdRS   = IfId_ir[25:21]; // rs field of instruction    
    wire [4:0] IdRT   = IfId_ir[20:16]; // rt field of instruction
    wire [4:0] IdRD   = IfId_ir[15:11]; // rd field of instruction
    wire [5:0] IdFCT  = IfId_ir[5:0];   // funct field of instruction   
    wire [31:0] IdRRS, IdRRT;           // register value of rs and rt

    assign bstall = IdB && 
                  ((IdRS!=0 && (IdRS==IdEx_dst || IdRS==ExMa_dst)) ||
                   (IdRT!=0 && (IdRT==IdEx_dst || IdRT==ExMa_dst))); // branch stall
                   
    GPR gpr(.CLK(CLK), .REGNUM0(IdRS), .REGNUM1(IdRT), .DOUT0(IdRRS), .DOUT1(IdRRT), // register file
            .REGNUM2(ExMa_dst), .DIN0(MaRSLT), .WE0(!PSTALL));

    reg [6:0]   IdOPN;   // instruction operation number (unique operation ID)
    reg [4:0]   IdDST;   // destination register
    reg [`ATTR] IdATTR;  // instruction attribute
    always @(IfId_ir) begin
        IdOPN   = `ERROR____; 
        IdDST  = 0;  
        IdATTR = 0;
        case (IdOP) // OP
            6'h00: case (IdFCT) // FUNCT 
                6'h00: begin IdOPN= (IdRD) ? `SLL______ : `NOP______; IdDST=IdRD; end
                6'h02: begin IdOPN=`SRL______; IdDST=IdRD;                        end
                6'h03: begin IdOPN=`SRA______; IdDST=IdRD;                        end
                6'h04: begin IdOPN=`SLLV_____; IdDST=IdRD;                        end
                6'h06: begin IdOPN=`SRLV_____; IdDST=IdRD;                        end
                6'h07: begin IdOPN=`SRAV_____; IdDST=IdRD;                        end
                6'h08: begin IdOPN=`JR_______; IdDST=0;                           end
                6'h09: begin IdOPN=`JALR_____; IdDST=31;                          end
                6'h10: begin IdOPN=`MFHI_____; IdDST=IdRD;                        end
                6'h12: begin IdOPN=`MFLO_____; IdDST=IdRD;                        end
                6'h18: begin IdOPN=`MULT_____; IdDST=0;                           end
                6'h19: begin IdOPN=`MULTU____; IdDST=0;                           end
                6'h20: begin IdOPN=`ADD______; IdDST=IdRD;                        end
                6'h21: begin IdOPN=`ADDU_____; IdDST=IdRD;                        end
                6'h22: begin IdOPN=`SUB______; IdDST=IdRD;                        end
                6'h23: begin IdOPN=`SUBU_____; IdDST=IdRD;                        end
                6'h24: begin IdOPN=`AND______; IdDST=IdRD;                        end
                6'h25: begin IdOPN=`OR_______; IdDST=IdRD;                        end
                6'h26: begin IdOPN=`XOR______; IdDST=IdRD;                        end
                6'h27: begin IdOPN=`NOR______; IdDST=IdRD;                        end
                6'h2a: begin IdOPN=`SLT______; IdDST=IdRD;                        end
                6'h2b: begin IdOPN=`SLTU_____; IdDST=IdRD;                        end
                6'h1a: begin IdOPN=`DIV______; IdDST=IdRD;                        end
                6'h1b: begin IdOPN=`DIVU_____; IdDST=IdRD;                        end
            endcase
            6'h01: case (IdRT)
                5'h00: begin IdOPN=`BLTZ_____; IdDST=0;                           end
                5'h01: begin IdOPN=`BGEZ_____; IdDST=0;                           end
            endcase
            6'h02:     begin IdOPN=`J________; IdDST=0;                           end
            6'h03:     begin IdOPN=`JAL______; IdDST=31;                          end
            6'h04:     begin IdOPN=`BEQ______; IdDST=0;                           end
            6'h05:     begin IdOPN=`BNE______; IdDST=0;                           end
            6'h06:     begin IdOPN=`BLEZ_____; IdDST=0;                           end
            6'h07:     begin IdOPN=`BGTZ_____; IdDST=0;                           end
            6'h08:     begin IdOPN=`ADDI_____; IdDST=IdRT;                        end
            6'h09:     begin IdOPN=`ADDIU____; IdDST=IdRT;                        end
            6'h0a:     begin IdOPN=`SLTI_____; IdDST=IdRT;                        end
            6'h0b:     begin IdOPN=`SLTIU____; IdDST=IdRT;                        end
            6'h0c:     begin IdOPN=`ANDI_____; IdDST=IdRT;                        end
            6'h0d:     begin IdOPN=`ORI______; IdDST=IdRT;                        end
            6'h0e:     begin IdOPN=`XORI_____; IdDST=IdRT;                        end
            6'h0f:     begin IdOPN=`LUI______; IdDST=IdRT;                        end
            6'h23:     begin IdOPN=`LW_______; IdDST=IdRT; IdATTR=`LD_4B;         end
            6'h2b:     begin IdOPN=`SW_______; IdDST=0;    IdATTR=`ST_4B;         end
        endcase
    end
    
    wire [`ADDR] IdBPC = IfId_npc + ({{16{IfId_ir[15]}}, IfId_ir[15:0]} << 2);
    always @(*) begin  ////// branch & jump resolution unit
        {IdTPC, IdC, IdB} = 0;
        case (IdOP)
            6'h04: begin IdB=1; IdTPC = IdBPC; IdC = (IdRRS == IdRRT);         end        // BEQ
            6'h05: begin IdB=1; IdTPC = IdBPC; IdC = (IdRRS != IdRRT);         end        // BNE
            6'h06: begin IdB=1; IdTPC = IdBPC; IdC = ( IdRRS[31]||(IdRRS==0)); end        // BLEZ
            6'h07: begin IdB=1; IdTPC = IdBPC; IdC = (~IdRRS[31]&&(IdRRS!=0)); end        // BGTZ
            6'h01: begin IdB=1; IdTPC = IdBPC; IdC = (IdRT) ? ~IdRRS[31] : IdRRS[31]; end // BGEZ,BLTZ
            6'h02: begin IdB=1; IdTPC = {IfId_npc[31:28], IfId_ir[25:0], 2'b0}; IdC=1;end // J
            6'h03: begin IdB=1; IdTPC = {IfId_npc[31:28], IfId_ir[25:0], 2'b0}; IdC=1;end // JAL
            6'h00: if      (IdFCT==6'h08) begin IdB=1; IdTPC = IdRRS; IdC = 1; end        // JR
                   else if (IdFCT==6'h09) begin IdB=1; IdTPC = IdRRS; IdC = 1; end        // JALR        
        endcase
    end
    
    always @(posedge CLK or negedge RST_X) begin ///// update pipeline registers
        if(!RST_X) {IdEx_npc, IdEx_rrs, IdEx_rrt, IdEx_dst, IdEx_ir, IdEx_opn, IdEx_attr} <= 0;
        else if(!PSTALL) begin
            IdEx_npc   <= (bstall) ? 0 : IfId_npc;
            IdEx_rrs   <= (bstall) ? 0 : IdRRS;    // data from general-purpose register file
            IdEx_rrt   <= (bstall) ? 0 : IdRRT;    // data from general-purpose register file
            IdEx_dst   <= (bstall) ? 0 : IdDST;
            IdEx_ir    <= (bstall) ? 0 : IfId_ir;
            IdEx_opn   <= (bstall) ? 0 : IdOPN;
            IdEx_attr  <= (bstall) ? 0 : IdATTR;
        end
    end
    
    /**************************************************************************************************/
    /* Stage 3 : EX, execute & address generation for LD/ST                                           */
    /**************************************************************************************************/
    wire        [31:0] RRS_U, RRT_U;                       // output of data forwarding unit
    wire signed [31:0] RRS_S = RRS_U;                      // signed wire
    wire signed [31:0] RRT_S = RRT_U;                      // signed wire
    wire         [4:0] SHAMT  = IdEx_ir[10:6];             // Shift amount
    wire        [15:0] IMM    = IdEx_ir[15:0];             // Immediate
    wire        [31:0] SET32I = {{16{IMM[15]}}, IMM};      // Sign Extended Imm.
    wire       [`ADDR] SETADI = SET32I[`ADDR] << 2;        // shifted immediate of address bit width
    wire       [`ADDR] JADDR  = IdEx_ir[`ADDR] << 2;       // Juma address
    wire       [`ADDR] ExA = RRS_U[`ADDR] + SET32I[`ADDR]; // mem address of LD/ST
    reg         [31:0] ExRSLT;                             // execution result
    reg          [3:0] ExWE;                               // memory write enable vector

    wire DIVINIT  = IdOPN==`DIV______  || IdOPN==`DIVU_____;
    wire EXSIGNED = IdEx_opn==`DIV______;
    wire [63:0] DURSLT;
    
    DIVUNIT divu(.CLK(CLK), .RST_X(RST_X), .INIT(DIVINIT), .SIGNED(EXSIGNED),
                 .A(RRS_U), .B(RRT_U), .RSLT(DURSLT), .BUSY(DIVBUSY));
    
    FORWARDING frs(.SRC(IdEx_ir[25:21]), .DIN0(IdEx_rrs), .DOUT(RRS_U), 
                   .DST1(ExMa_dst), .DIN1(ExMa_rslt), .DST2(MaWb_dst), .DIN2(MaWb_rslt));
    FORWARDING frt(.SRC(IdEx_ir[20:16]), .DIN0(IdEx_rrt), .DOUT(RRT_U), 
                   .DST1(ExMa_dst), .DIN1(ExMa_rslt), .DST2(MaWb_dst), .DIN2(MaWb_rslt));

    always @(*) begin
        {ExRSLT, ExWE} = 0;
        case ( IdEx_opn )
          `ADD______ : begin ExRSLT = RRS_U + RRT_U;                                             end
          `ADDI_____ : begin ExRSLT = RRS_U + SET32I;                                            end
          `ADDIU____ : begin ExRSLT = RRS_U + SET32I;                                            end
          `ADDU_____ : begin ExRSLT = RRS_U + RRT_U;                                             end
          `SUB______ : begin ExRSLT = RRS_U - RRT_U;                                             end
          `SUBU_____ : begin ExRSLT = RRS_U - RRT_U;                                             end
          `AND______ : begin ExRSLT = RRS_U & RRT_U;                                             end
          `ANDI_____ : begin ExRSLT = RRS_U & {16'h0, IMM};                                      end
          `NOR______ : begin ExRSLT = ~(RRS_U | RRT_U);                                          end
          `OR_______ : begin ExRSLT = RRS_U | RRT_U;                                             end
          `ORI______ : begin ExRSLT = RRS_U | {16'h0, IMM};                                      end
          `XOR______ : begin ExRSLT = RRS_U ^ RRT_U;                                             end
          `XORI_____ : begin ExRSLT = RRS_U ^ {16'h0, IMM};                                      end
          `SLL______ : begin ExRSLT = RRT_U <<  SHAMT;                                           end
          `SRL______ : begin ExRSLT = RRT_U >>  SHAMT;                                           end
          `SRA______ : begin ExRSLT = RRT_S >>> SHAMT;                                           end
          `SLLV_____ : begin ExRSLT = RRT_U <<  RRS_U[4:0];                                      end
          `SRLV_____ : begin ExRSLT = RRT_U >>  RRS_U[4:0];                                      end
          `SRAV_____ : begin ExRSLT = RRT_S >>> RRS_U[4:0];                                      end
          `SLT______ : begin ExRSLT = (RRS_U[31] ^ RRT_U[31]) ? RRS_U[31] : (RRS_U < RRT_U);     end
          `SLTI_____ : begin ExRSLT = (RRS_U[31] ^   IMM[15]) ? RRS_U[31] : (RRS_U < SET32I);    end
          `SLTIU____ : begin ExRSLT = (RRS_U < SET32I);                                          end
          `SLTU_____ : begin ExRSLT = (RRS_U < RRT_U);                                           end
          `JAL______ : begin ExRSLT = IdEx_npc + 4;                                              end
          `JALR_____ : begin ExRSLT = IdEx_npc + 4;                                              end
          `LUI______ : begin ExRSLT = {IMM, 16'h0};                                              end
          `LW_______ : begin ExRSLT = {ExA[31:2], 2'b00};                                        end
          `SW_______ : begin ExRSLT = {ExA[31:2], 2'b00}; ExWE = 4'b1111;                        end
          `MFHI_____ : begin ExRSLT = hi;                                                        end
          `MFLO_____ : begin ExRSLT = lo;                                                        end
        endcase
    end

    always @( posedge CLK or negedge RST_X ) begin  ///// update hi and lo register
        if(!RST_X) {hi, lo} <= 0;
        else if(!PSTALL) begin
            if(IdEx_opn == `MULT_____) {hi, lo} <= RRS_S * RRT_S;
            if(IdEx_opn == `MULTU____) {hi, lo} <= RRS_U * RRT_U;
            if(IdEx_opn == `DIV______) {hi, lo} <= (RRT_U) ? DURSLT : 0;
            if(IdEx_opn == `DIVU_____) {hi, lo} <= (RRT_U) ? DURSLT : 0;             
        end
    end

    always @( posedge CLK or negedge RST_X ) begin ///// update pipeline registers
        if(!RST_X) {ExMa_rslt, ExMa_dst, ExMa_mwe, ExMa_oe, ExMa_lds, ExMa_std} <= 0;
        else if(!PSTALL) begin
            ExMa_rslt <= ExRSLT;
            ExMa_dst  <= IdEx_dst;
            ExMa_std  <= RRT_U;
            ExMa_mwe  <= ExWE;
            ExMa_oe   <= (IdEx_attr & `LDST_ANY) ? 1 : 0;
            ExMa_lds  <= (IdEx_opn==`LW_______) ?  1 : 0;
        end 
    end

    /**************************************************************************************************/
    /* Stage 4 : MA, memory access for LD/ST                                                          */
    /**************************************************************************************************/
    assign D_ADDR = (ExMa_oe)            ? ExMa_rslt : 0;
    assign D_OUT  = (ExMa_oe)            ? ExMa_std  : 0;
    assign D_WE   = (ExMa_oe && !PSTALL) ? ExMa_mwe  : 0;
    assign D_OE   =  ExMa_oe;

    // wire [31:0] MaLD_ = (ExMa_rslt[23]) ? IO_IN : D_IN; // use I/O input if the high address
    wire [31:0] MaLD_ = D_IN; // This edition does not use I/O.
    wire [31:0] MaLDD = MaLD_ >> (8*ExMa_rslt[1:0]);    // loaded data, align data here

    assign MaRSLT = (ExMa_lds) ? MaLDD : ExMa_rslt;

    always @( posedge CLK or negedge RST_X ) begin ///// update pipeline registers
        if(!RST_X) {MaWb_rslt, MaWb_dst} <= 0;
        else if(!PSTALL) begin
            MaWb_rslt <= MaRSLT;
            MaWb_dst  <= ExMa_dst;
        end
    end

    /* Update STATUS_LED (ERROR_LED)                                                                  */
    /**************************************************************************************************/
    wire [`OPNT] ExOPN = IdEx_opn;  
    always @( posedge CLK or negedge RST_X ) begin
        if(!RST_X) STAT <= 0;
        else if(!PSTALL) begin
            if(IdOPN==`ERROR____) STAT[0] <= 1; // instruction decode error!
            if(((ExOPN==`LW_______ || ExOPN==`SW_______) && ExA[1:0]!=0))
               STAT[1] <= 1; // memory access alignment error!
        end
    end
endmodule

/******************************************************************************************************/
/* 32bitx32 2R/1W General Purpose Registers (Register File)                                           */
/******************************************************************************************************/
module GPR(CLK, REGNUM0, REGNUM1, REGNUM2, DIN0, WE0, DOUT0, DOUT1);
    input             CLK;
    input       [4:0] REGNUM0, REGNUM1, REGNUM2;
    input      [31:0] DIN0;
    input             WE0;
    output reg [31:0] DOUT0, DOUT1;

    reg [31:0] r[0:31];

    always @(negedge CLK) DOUT0 <= (REGNUM0==0) ? 0 : r[REGNUM0];
    always @(negedge CLK) DOUT1 <= (REGNUM1==0) ? 0 : r[REGNUM1];
    always @(posedge CLK) if(WE0) r[REGNUM2] <= DIN0;
endmodule

/***** data forwarding unit                                                                       *****/
/******************************************************************************************************/
module FORWARDING(SRC, DIN0, DST1, DIN1, DST2, DIN2, DOUT);
    input   [4:0] SRC, DST1, DST2;   // register number 
    input  [31:0] DIN0, DIN1, DIN2;  // 32bit value
    output [31:0] DOUT;              // 32bit data output

    assign DOUT = (SRC!=0 && DST1==SRC) ? DIN1 : (SRC!=0 && DST2==SRC) ? DIN2 : DIN0;
endmodule

/******************************************************************************************************/
