/******************************************************************************************************/
/* MieruPC 2010: Main Memory (SRAM Controller)                                    ArchLab. TOKYO TECH */
/*  See the README file on the parent directory for license information.                              */
/******************************************************************************************************/
`include "../rtl/define.v"
`default_nettype none

/******************************************************************************************************/
module IMEM(CLK, RST_X, A_IN, D_IN, WE_IN, INST_ADDR, INST_OUT);
    input         CLK, RST_X;
    input  [31:0] A_IN;
    input  [31:0] D_IN;
    input         WE_IN;
    input  [31:0] INST_ADDR;
    output [31:0] INST_OUT;
    
    reg [31:0] DATA, waddr;
    reg        WE;
    
    wire [31:0] a = (waddr<`IMEM_SIZE/4) ? A_IN : INST_ADDR;
    IMEM_CORE(CLK, WE, a, DATA, INST_OUT); 
    
    always @(posedge CLK or negedge RST_X) begin
        if(!RST_X) begin
            {DATA, WE, waddr} <= 0;
        end else begin
            if(WE_IN) begin
                WE   <= (waddr <`IMEM_SIZE/4) ? 1 : 0; // note
                DATA <= D_IN;
                waddr <= waddr + 1;
            end else begin
                WE <= 0;
            end
        end
    end
endmodule

/******************************************************************************************************/
module IMEM_CORE(CLK, WE_IN, A_IN, D_IN, D_OUT);
    input             CLK, WE_IN;
    input [31:0]      A_IN, D_IN;
    output reg [31:0] D_OUT;
      
    reg [31:0]  mem[`IMEM_SIZE/4-1:0];

    always @(negedge CLK) begin // use negedge CLK
        if(WE_IN) mem[A_IN[15:2]] <= D_IN;
        D_OUT <=  mem[A_IN[15:2]];
    end
endmodule
/******************************************************************************************************/
